import java.beans.*;
import java.lang.reflect.*;
import java.util.*;

class Main {
    public static void main(String[] args) {
        if (args.length < 1) {
            System.err.println("Usage: java Main <name of bean class file>");
            System.exit(1);
        }

        // The other args are to be appended to the search path.
        String[] oldPath = Introspector.getBeanInfoSearchPath();
        String[] newPath = new String[oldPath.length + args.length - 1];
        for (int i=0; i<oldPath.length; i++) {
            newPath[i] = oldPath[i];
        }
        for (int i=0; i<args.length-1; i++) {
            newPath[i+oldPath.length] = args[i+1];
        }
        Introspector.setBeanInfoSearchPath(newPath);

        try {
            // Get the bean.
            Class bc = FileClassLoader.load(args[0]);

            printBeanInfo(bc);
        } catch (IntrospectionException e) {
            e.printStackTrace();
        }
    }

    public static void printBeanInfo(Class c) throws IntrospectionException {
        BeanInfo bi = Introspector.getBeanInfo(c);
        BeanDescriptor bd = bi.getBeanDescriptor();

        System.out.println("Class: " + bd.getBeanClass().getName());
        System.out.println("Name: " + bd.getName());
        System.out.println("Display name: " + bd.getDisplayName());
        System.out.println("Short Description: " + bd.getShortDescription());
        System.out.println("Expert: " + bd.isExpert());
        System.out.println("Hidden: " + bd.isHidden());
        if (bd.getCustomizerClass() != null) {
            System.out.println("Customizer: " + 
                bd.getCustomizerClass().getName());
        }
        System.out.println();

        // Print legend
        System.out.println(
          "h:hidden, e:expert, i:indexed, *:default, c:constrained, b:bound");
        System.out.println("u:unicast, d:default set\n");

        // ** Print the property descriptors.
        printRepeat(" ", 30);
        System.out.println("PROPERTIES");
        printRepeat("-", 72);
        System.out.println();
        print("Attr", 5);
        print("Name", 20);
        print("Type", 25);
        print("Accessor methods", 25);
        System.out.println();
        printRepeat("-", 72);
        System.out.println();

        // Now print each descriptor.
        PropertyDescriptor pds[] = bi.getPropertyDescriptors();
        for (int i=0; i<pds.length; i++) {
            printProperty(pds[i], i == bi.getDefaultPropertyIndex());
        }
        System.out.println();

    // ** Print the event set descriptors.
        printRepeat(" ", 30);
        System.out.println("EVENT SETS");
        printRepeat("-", 72);
        System.out.println();
        print("Attr", 5);
        print("Name", 20);
        System.out.println("Registration methods");
        System.out.println("        Callback methods");
        printRepeat("-", 72);
        System.out.println();

        // Now print each event set descriptor.
        EventSetDescriptor esds[] = bi.getEventSetDescriptors();
        for (int i=0; i<esds.length; i++) {
            printEventSet(esds[i], i == bi.getDefaultEventIndex());
        }
        System.out.println();

    // ** Print the method descriptors.
        printRepeat(" ", 30);
        System.out.println("METHODS");
        printRepeat("-", 72);
        System.out.println();
        print("Attr", 5);
        System.out.println("Signature");
        printRepeat("-", 72);
        System.out.println();

        // Now print each method descriptor.
        MethodDescriptor methods[] = bi.getMethodDescriptors();
        for (int i=0; i<methods.length; i++) {  
            printMethods(methods[i]);
        }
    }

    static void printProperty(PropertyDescriptor pd, boolean def) {
        printAttributes(pd, def);
        print(pd.getName(), 20);
        printName(pd.getPropertyType(), 25);

        // Print accessor methods.
        printName(pd.getReadMethod(), -1);
        System.out.print("/");
        printName(pd.getWriteMethod(), -1);
        System.out.println();

        if (pd instanceof IndexedPropertyDescriptor) {
            IndexedPropertyDescriptor ipd = (IndexedPropertyDescriptor)pd;

            printAttributes(ipd, def, true);
            print("", 20);
            printName(ipd.getIndexedPropertyType(), 25);
            printName(ipd.getIndexedReadMethod(), -1);
            System.out.print("/");
            printName(ipd.getIndexedWriteMethod(), -1);
            System.out.println();
        }
    }

    static void printEventSet(EventSetDescriptor esd, boolean def) {
        printAttributes(esd, def);
        printName(esd, 20);

        // Print listener methods.
        printName(esd.getAddListenerMethod(), -1);
        System.out.print("/");
        printName(esd.getRemoveListenerMethod(), -1);
        System.out.println();

        // Print callback methods.
        Method ems[] = esd.getListenerMethods();
        for (int j=0; j<ems.length; j++) {
            System.out.print("        ");
            printName(ems[j], -1);
            System.out.println();
       }
    }

    static void printMethods(MethodDescriptor md) {
        printAttributes(md, false);
        System.out.println(md.getMethod().toString());

        if (md.getShortDescription() != null 
                && !md.getShortDescription().equals(md.getName())) {
            System.out.println(".....\"" + md.getShortDescription() + "\"");
        }

        ParameterDescriptor[] pds = md.getParameterDescriptors();
        if (pds != null) {
            for (int i=0; i<pds.length; i++) {
                System.out.print(".....");
                printAttributes(pds[i], false);
                print(pds[i].getName(), 10);
                System.out.println();
                if (pds[i].getShortDescription() != null) {
                    System.out.println(".....     \"" +
                        pds[i].getShortDescription()+"\"");
                }
            }
        }
    }

    // Prints s and pads the string with Math.max(1, width-s.length()) spaces.
    static void print(String s, int width) {
        System.out.print(s);
        for (int i=0; i<Math.max(1, (width-s.length())); i++) {
            System.out.print(" ");
        }
    }

    // Prints s 'count' times.
    static void printRepeat(String s, int count) {
        for (int i=0; i<count; i++) {
            System.out.print(s);
        }
    }

    // Prints the object's name.  If null, prints "?".
    // Also if width >= 0, pads the string with 
    // Math.max(1, width-s.length()) spaces.
    static void printName(Object o, int width) {
        String result;

        if (o == null) {
            result = "?";
        } else if (o instanceof Class) {
            result = ((Class)o).getName();
        } else if (o instanceof Method) {
            result = ((Method)o).getName();
        } else if (o instanceof FeatureDescriptor) {
            result = ((FeatureDescriptor)o).getName();
        } else {
            throw (new RuntimeException());
        }
        if (width >= 0) {
            print(result, width);
        } else {
            System.out.print(result);
        }
    }

    static void printAttributes(FeatureDescriptor fd, boolean def) {
        printAttributes(fd, def, false);
    }
    static void printAttributes(FeatureDescriptor fd, boolean def, 
                                boolean indexed) {
        String result = "";

        if (def) {
            result += "*";
        } 
        if (fd.isHidden()) {
            result += "h";
        }
        if (fd.isExpert()) {
            result += "e";
        }
        if (indexed) {
            result += "i";
        }
        if (fd instanceof PropertyDescriptor) {
            PropertyDescriptor pd = (PropertyDescriptor)fd;

            if (pd.isConstrained()) {
                result += "c";
            }
            if (pd.isBound()) {
                result += "b";
            }
        }
        if (fd instanceof EventSetDescriptor) {
            EventSetDescriptor esd = (EventSetDescriptor)fd;

            if (esd.isUnicast()) {
                result += "u";
            }
            if (esd.isInDefaultEventSet()) {
                result += "d";
            }
        }
        print(result, 5);
    }
}
